#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
# Initial Release:
# A two stage process with options to create characters from a custom font and then (with a user added path) add lots of layers for a 3D trail effect
#
# Version 0.02 added Normal mode with hue darkening and No darkening to list of modes and darkened layers using hue saturation filter
#              added trim option for first character copies
#			   added option to update screen display (which also freezes layer dock update)
#
# Version 0.03 added options for rotating characters along the user path
#              by angle option rotates characters by smaller increments (0.1 degrees)
#              by turn option rotates characters by increments of quarter turns up to two whole turns
#              added auto blur option (in addition to user blur) to control the blurring of the perspective trail
#              added option to scale the grabbed characters image (up 75% of the original)
#              added option to set a limit on the length of the path
#
# It can be executed by selecting the menu option: 'Filters>Custom Font Tools>>Custom Font 3D Wordy v x.xx


from gimpfu import *
import math	
#from array import array
#import math
#import random

#from gtk import gdk
#w = gdk.screen_width()
#h = gdk.screen_height()

# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)

#####
##### TEMP VALUES

# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)
darkestgrey = gimpcolor.RGB(20, 20, 20)
verydarkgrey = gimpcolor.RGB(50, 50, 50)
darkgrey = gimpcolor.RGB(75, 75, 75)
midgrey2 = gimpcolor.RGB(105, 105, 105)
midgrey = gimpcolor.RGB(125, 125, 125)
lightgrey = gimpcolor.RGB(170, 170, 170)
verylightgrey = gimpcolor.RGB(222, 222, 222)
red = gimpcolor.RGB(220,10,10)	



def do_stuff(image, layer, word_text, grab, do_pre_merge, blur, mode, final_height, rotate_style, distance_step, angle_inc, num_turns, clockwise, scale_value, trim, do_flush, do_merge, shorten, autoblur):
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)	
	
    apply_to_custom = True
    dummy = 0
    active_layer_id = pdb.gimp_image_get_active_layer(image)
    num_turns = float(num_turns*0.25)
    if angle_inc>0.0 or num_turns>0.0: do_follow = True
    shorten = (100-shorten)
	
    if clockwise ==0:
        clockwise = True
    else:
        clockwise = False
	
    loop_ptr = 1
    first_char_layer_name = "none"
	
    if grab: # grab the characters from the opened custom font
	
        # get lists of vector ids in image
        num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
	
        # get lists of layers etc in image
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
        original_active_layer = image.active_layer
		
        num_multi_layers,other_layers,num_char_layers = 0,0,0
        visi_loop, first_char_layer  = 0, -1
    
        do_wordy, loop, test_exists, new_word = True, 0, True, ""
        while loop<len(word_text): #check that layers exist
            test_layer = word_text[loop]
            test_char = test_layer
            if test_layer == " ": 
                test_layer = "space"
                test_char = " "
            test_layer = test_layer + ".png"
            check_layer = pdb.gimp_image_get_layer_by_name(image, test_layer)
            if check_layer<0:
                test_char = test_layer[0]
                if test_char.isupper():
                    test_char = test_char.lower()
                    test_layer = test_char + ".png"
                    check_layer = pdb.gimp_image_get_layer_by_name(image, test_layer)
                    if check_layer<0: 
                        test_exists = False
                    else:
                        new_word += test_char
                elif test_char.islower():
                    test_char = test_char.upper()
                    test_layer = test_char + ".png"
                    check_layer = pdb.gimp_image_get_layer_by_name(image, test_layer)
                    if check_layer<0: 
                        test_exists = False
                    else:
                        new_word +=  test_char
            else:
                new_word +=  test_char
            loop +=1
        if test_exists == False: do_wordy = False
	
        word_text = new_word
	
        if do_wordy:
            perspective_image = pdb.gimp_image_new(image.width, image.height, 0) # add a new image for the perspective word
            loop, word_x, kerning_value = 0, 0, 0
			
            while loop < len(word_text): # copy the characters from the custom font image file to the new perspective image
                if str(word_text[loop]) == " ":
                    perspective_layer_name = "space.png"
                else:
                    perspective_layer_name = str(word_text[loop])+".png"

                temp_layer_id = pdb.gimp_image_get_layer_by_name(image, perspective_layer_name) # get layer
                perspective_layer_id = pdb.gimp_layer_new_from_drawable(temp_layer_id, perspective_image) # copy it
                pdb.gimp_image_insert_layer(perspective_image, perspective_layer_id, None, loop) # add layer and ensure layer order is the same as in the perspective word
                pdb.gimp_item_set_visible(perspective_layer_id, True)
                temp_layer_copy = perspective_layer_id.copy(False) # make a copy of the layer
                perspective_image.add_layer(temp_layer_copy,0)
                perspective_image.remove_layer(temp_layer_copy) # remove temporary layer
				
                if loop >0:
                    pdb.gimp_item_transform_translate(perspective_layer_id, word_x, 0) # move layer to position in new image
					
                if loop<len(word_text)-1:
                    kerning_pair = word_text[loop] + word_text[loop + 1] +".png" # create a kerning path name
                    kern_vector = pdb.gimp_image_get_vectors_by_name(image,str(kerning_pair) ) # and see if it exists in the parent image
                    if kern_vector<0:
                        kerning_value = perspective_layer_id.width # ie no kerning path available so use character width for spacing
                    else:
                        NumberOfStrokes_in_Path,ActivePath_StrokeIDs_List = pdb.gimp_vectors_get_strokes(kern_vector) # get vector information for kerning path width
                        ActivePath_StrokeID = ActivePath_StrokeIDs_List[0]
                        kerning_value = pdb.gimp_vectors_stroke_get_length(kern_vector,ActivePath_StrokeID,0.01) # ie the length of the existing kerning path
					
                word_x += kerning_value # update spacing
                if word_text[loop] == " ": pdb.gimp_image_remove_layer(perspective_image, perspective_layer_id) # remove original space layer from image
                loop += 1
			
            pdb.gimp_image_resize_to_layers(perspective_image) # resize image to accommodate layers
            perspective_display = pdb.gimp_display_new(perspective_image) # display the perspective word characters
			
            if do_pre_merge: 
                temp_layer_copy = pdb.gimp_image_merge_visible_layers(perspective_image, 0) # merge all the copies of the character if required				
				
            if scale_value>0:
                old_interpolation = pdb.gimp_context_get_interpolation() # Note current interpolation method
                pdb.gimp_context_set_interpolation(3) # Ensure that interpolation is set to required method
                scale_value = 100-scale_value
                new_image_width =  (float(perspective_image.width/100.0))*float(scale_value)
                new_image_height =  float(perspective_image.height*float(new_image_width/perspective_image.width))
                pdb.gimp_image_scale(perspective_image, new_image_width, float(perspective_image.height*float(new_image_width/perspective_image.width)) )
                pdb.gimp_context_set_interpolation(old_interpolation) # Restore old interpolation method			
			
        else:
            message_text = "There appears to be no custom font layers associated with this image or the required character layers are missing from this file."
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message(message_text)
            pdb.gimp_message_set_handler(old_handler)    
	
    else: # start drawing grabbed characters along a user path
	
        # get lists of vector ids in image
        num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
	
        # get lists of layers in image
        num_persp_layers, persp_layer_ids = pdb.gimp_image_get_layers(image) # get info on  layers
        persp_layer_names_list = image.layers
        original_active_layer = image.active_layer
		
        if num_vectors<num_persp_layers: # check that there are enough paths available
            message_text = "There are not enough paths to continue. Please add one path for each character."
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message(message_text)
            pdb.gimp_message_set_handler(old_handler)   

        else:
		
            new_image_height = image.height
            loop = 0
			
            while loop<num_persp_layers:
                pdb.gimp_item_set_visible(persp_layer_names_list[loop], False)
                loop +=1
		
            # set mode for 3D layers (generally increasingly darker)
            if mode < 2:
                darken_mode = 28 # Normal 
            elif mode == 2:
                darken_mode = 35 # darken only
            elif mode == 3:
                darken_mode = 54 # Luma/luminance darken
            elif mode == 4:
                darken_mode = 30 # Multiply
            elif mode == 5:
                darken_mode = 43 # Burn
            elif mode == 6:
                darken_mode = 53 # Linear burn
			
            persp_loop = 0
            while persp_loop<num_persp_layers: # loop through available character layers
                persp_item = persp_loop
                current_vector = vector_list[persp_item] # get current vector
		
                NumberOfStrokes_in_Path,ActivePath_StrokeIDs_List = pdb.gimp_vectors_get_strokes(current_vector)
                ActivePath_StrokeID = ActivePath_StrokeIDs_List[0]
                path_length = pdb.gimp_vectors_stroke_get_length(current_vector,ActivePath_StrokeID,0.01)
                persp_layer = persp_layer_names_list[persp_item] # current layer
                original_name = pdb.gimp_item_get_name(persp_layer) # make a note of the original layer name
                original_position = pdb.gimp_image_get_item_position(image, persp_layer) # make a note of the original layer position
                persp_height = persp_layer.height
                persp_width = persp_layer.width
                scaled_scale = float(0 - (persp_height-final_height)/path_length) # set scale value according to path length and the final layer height set by user
				
                loop = 0
                dist = 0
                word_x = 0
                temp_layer = pdb.gimp_layer_copy(persp_layer, True) # copy and insert layer
                my_darkness = float(100/path_length)
                my_grey = float(0.7/path_length) # set range for shading to reduce use of too many very black shades
                new_grey = 1.0 # set initial shading to white
                
                grey_shade = gimpcolor.RGB(my_grey, my_grey, my_grey) # colour value
		
                if darken_mode==28 and do_flush: pdb.gimp_image_freeze_layers(image)
                angle = 0.0
                target = path_length 
                if shorten>0: # reset target if required by user
                    target = float(path_length/100)*shorten
					
                angle_blur = (new_image_height//50)+blur # angle_blur used as autoblur
                if angle_blur<0: angle_blur = 0
				
                if rotate_style == 1:
                    angle_inc = (360.0/path_length)*num_turns
                pdb.gimp_context_set_feather(True) # aim to set a slight feathering for edges of selections
				
                while  dist<target:
                    grey_shade = gimpcolor.RGB(new_grey, new_grey, new_grey) # set colour to use in fill layer before applying with user mode
                    pdb.gimp_context_set_foreground(grey_shade)
			
                    temp_layer_position = pdb.gimp_image_get_item_position(image, persp_layer) # find position of layer in the item tree
                    temp_layer2 = pdb.gimp_layer_copy(persp_layer, True) # copy and insert layer
                    pdb.gimp_image_insert_layer(image, temp_layer2, None, 0)
                    pdb.gimp_item_set_visible(temp_layer2, True)
                    if darken_mode<>28:
                        fill_layer2 = gimp.Layer(image, "Fill Layer2", persp_layer.width, persp_layer.height, RGBA_IMAGE, 100, darken_mode) # add a fill layer with selected darker mode
                        pdb.gimp_drawable_fill(fill_layer2, 0) # fill layer with current shade of grey
                        image.add_layer(fill_layer2, -1)
                        pdb.gimp_item_set_visible(temp_layer2, True) # Make sure layer copy is visible
                    else: # normal mode used
                        temp_layer = temp_layer2
                        if mode==1: pdb.gimp_drawable_hue_saturation(temp_layer, 0, 0, 0-((loop*distance_step)*my_darkness), 0, 0)
                    pdb.gimp_item_set_name(temp_layer,"copy layer")
                    if loop>0:
                        pdb.gimp_context_set_feather_radius(2, 2) # slight feather for edges
                        if autoblur:
                            if angle_blur>0: pdb.plug_in_gauss(image, temp_layer, angle_blur, angle_blur, 0)
                        else:
                            if blur>0: 
                                pdb.plug_in_gauss(image, temp_layer, blur, blur, 0)
                                if ((path_length-loop) % 100) == 0: # alters user blur after set number of layers added
                                    blur -= 1
                                    if blur<0: blur = 0
                                
                        if angle_blur > 0:
                            if ((path_length-loop) % 100) == 0: # alters autoblur after set number of layers added
                                    
                                angle_blur -= 1
                        if angle_blur<0: angle_blur = 0
						
                    else:
                        if trim>0: proc_trim_away(image, temp_layer, trim) # trim first copy by UI amount
						
						
                    if darken_mode<>28:
                        temp_layer = pdb.gimp_image_merge_down(image, fill_layer2, 2) #  and merge down the fill layer onto the layer copy
                        pdb.gimp_layer_set_mode(temp_layer, 28) # set mode to normal
		
		
                    pdb.gimp_layer_scale(temp_layer, temp_layer.width*(float(temp_layer.height+(float(scaled_scale*(loop*distance_step))))/  temp_layer.height), temp_layer.height+(float((scaled_scale*(loop*distance_step)))), False)# scale by height rather than by width
                    persp_height = temp_layer.height
			
                    x_point,y_point,SlopeAtThisPoint,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(current_vector,ActivePath_StrokeID,dist,0.01) # get coordinates of point along path
                    x_point2,y_point2,SlopeAtThisPoint2,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(current_vector,ActivePath_StrokeID,dist+0.5,0.01)
										
                    translate_x = x_point
                    translate_y = y_point
					
                    pdb.gimp_layer_set_offsets(temp_layer, translate_x-(float(temp_layer.width)/2), translate_y-(float(temp_layer.height)/2) ) # position layer
					
					
                    if rotate_style<2: 
                        temp_layer = pdb.gimp_item_transform_rotate(temp_layer, math.radians(angle),True,translate_x-(float(temp_layer.width)/2),translate_y-(float(temp_layer.height)/2)) # rotate original layer on path
                        
					
                    pdb.gimp_image_lower_item_to_bottom(image, temp_layer)
                    if do_merge: temp_layer = pdb.gimp_image_merge_visible_layers(image, 0) # merge copies of the character
                    if do_flush: # update screen display
                        pdb.gimp_image_resize_to_layers(image)
                        success = pdb.gimp_displays_flush()
						
                    word_x += distance_step # update spacing
					
                    if rotate_style<2: # for rotation styles
                        if clockwise:
                            angle += (angle_inc*distance_step)
                        else:
                            angle -= (angle_inc*distance_step)
							
                    dist =  word_x # update spacing
                    new_grey = 1.0 - ((float(0.7/path_length))*(loop*distance_step)) # update fill colour value
                    loop += 1
			
                if darken_mode==28 and do_flush: pdb.gimp_image_thaw_layers(image)
                
                pdb.gimp_image_remove_layer(image, persp_layer_names_list[persp_item]) # remove original layer from image]
                pdb.gimp_image_reorder_item(image, temp_layer, None, original_position) # put the new layer into the position of the old layer
				
                pdb.gimp_item_set_visible(temp_layer, False)
                final_name = original_name + " x" + str(loop-1)
                pdb.gimp_item_set_name(temp_layer, final_name) # set a name to record the number of layers used to create the 3D effect
                if not do_merge: pdb.gimp_image_lower_item_to_bottom(image, temp_layer)
            
                persp_loop += 1
            
            # get lists of layers in image
            num_persp_layers, persp_layer_ids = pdb.gimp_image_get_layers(image) # update list of layers
            persp_layer_names_list = image.layers
            loop = 0
            pdb.gimp_image_resize_to_layers(image)
            while loop<num_persp_layers:
                pdb.gimp_item_set_visible(persp_layer_names_list[loop], True)
                loop += 1
            
					
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
######################################################################################################################################

    #### Useful Functions
# -------------------------------------------------------------------------------------
# -------------------------------------------------------------------------------------


#######################################################################################################
def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value

######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	
######################################################################################################################################
		
def proc_trim_away(img, layer, amount):
    #### Trim off pixels from a layer
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
    #### End of Proc_trim_away
	
######################################################################################################################################
######################################################################################################################################
	
register(
    "python_fu_Custom_Font_3D_wordy_v_003",
    "Provides options for adding perspective to characters along a path",
    "Provides options for creating and manipulating characters with perspective along a path",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2022",
    "<Image>/Filters/Custom Font Tools/Custom Font 3D Wordy v 0.03...",
    "RGB*",
    [
    (PF_STRING, "perspective_string", "Text: ", "GIMP"),
    (PF_TOGGLE, "option", "Grab characters: ", 0),
    (PF_TOGGLE, "pre_merge", "Merge grabbed characters (before adding paths): ", 0),
    (PF_SPINNER, "blur", "Add extra blur: ", 0, (0, 30, 1)),
    (PF_OPTION, "mode_option", "Mode to use: ", 1, ("No darkening", "Darken by hue", "Darken only", "Luma darken", "Multiply", "Burn", "Linear burn")),
    (PF_SPINNER, "final_ht", "Character height at the end of the path: ", 50, (10, 100, 1)),
    (PF_OPTION, "r_style", "Rotation style: ", 2, ("Rotate by angle", "Rotate by turn","No rotation")),
    (PF_SPINNER, "step", "Step size along path: ", 1, (1, 100, 1)),
    (PF_SPINNER, "rotations", "Angle size (style 1): ", 0.5, (0.0, 22.5, 0.1)),
    (PF_OPTION, "turns", "Number of turns (style 2): ", 2, ("No turn","one quarter turn","one half turn", "one three quarter turn", "one full turn", "one and a quarter turns", "one and a half turns" , "one and a three quarter turns", "two full turns" )),
    (PF_OPTION, "direction", "Rotation direction: ", 0, ("Clockwise", "Anti-clockwise")),
    (PF_SPINNER, "scale", "Scale grabbed characters by %: ", 0, (0, 75, 1)),
    (PF_SPINNER, "trim", "Trim initial face ", 0, (0, 5, 1)),
    (PF_TOGGLE, "update_screen", "Update screen: ", 1),
    (PF_TOGGLE, "merge", "Merge layers: ", 1),
    (PF_SPINNER, "path_stop", "Shorten path by %: ", 0, (0, 90, 1)),
    (PF_TOGGLE, "auto_off", "Use autoblur: ", 1),
    ],
    [],
    do_stuff)

main()
